from dataclasses import dataclass
from typing import Tuple
from unittest.mock import Mock

import pytest

from doge_datagen import DataOnlineGenerator, SubjectFactory


@dataclass
class SomeSubject:
    subject_id: int

    def __hash__(self):
        return self.subject_id


STATE_1 = 'state1'
STATE_2 = 'state2'
STATE_3 = 'state3'
TRANSITION_1 = 'transition1'
TRANSITION_2 = 'transition2'
SOME_SUBJECT_1 = SomeSubject(1)
SOME_SUBJECT_2 = SomeSubject(2)


class TestDoge:
    
    def test_should_transition_subjects(self):
        factory = self.__create_factory_mock([SOME_SUBJECT_1, SOME_SUBJECT_2])
        action_callback = Mock(return_value=True)
        event_sink = Mock()
        doge = DataOnlineGenerator([STATE_1, STATE_2, STATE_3],
                                   STATE_1,
                                   factory,
                                   2, 1000, 2)
        doge.add_transition(TRANSITION_1, STATE_1, STATE_2, 100, action_callback, [event_sink])
        doge.add_transition(TRANSITION_2, STATE_2, STATE_3, 100, action_callback, [event_sink])

        doge.start()

        assert action_callback.call_count == 4
        self.__assert_action_called(action_callback.call_args_list[0].args,
                                    SOME_SUBJECT_1, TRANSITION_1, STATE_1, STATE_2)
        self.__assert_action_called(action_callback.call_args_list[1].args,
                                    SOME_SUBJECT_2, TRANSITION_1, STATE_1, STATE_2)
        self.__assert_action_called(action_callback.call_args_list[2].args,
                                    SOME_SUBJECT_1, TRANSITION_2, STATE_2, STATE_3)
        self.__assert_action_called(action_callback.call_args_list[3].args,
                                    SOME_SUBJECT_2, TRANSITION_2, STATE_2, STATE_3)
        assert event_sink.collect.call_count == 4
        self.__assert_event_sink_collect_called(event_sink.collect.call_args_list[0].args,
                                                0, SOME_SUBJECT_1, TRANSITION_1, STATE_1, STATE_2)
        self.__assert_event_sink_collect_called(event_sink.collect.call_args_list[1].args,
                                                0, SOME_SUBJECT_2, TRANSITION_1, STATE_1, STATE_2)
        self.__assert_event_sink_collect_called(event_sink.collect.call_args_list[2].args,
                                                1000, SOME_SUBJECT_1, TRANSITION_2, STATE_2, STATE_3)
        self.__assert_event_sink_collect_called(event_sink.collect.call_args_list[3].args,
                                                1000, SOME_SUBJECT_2, TRANSITION_2, STATE_2, STATE_3)
        event_sink.close.assert_called()

    def test_should_transition_to_final_state(self):
        factory = self.__create_factory_mock([SOME_SUBJECT_1])
        action_callback = Mock(return_value=True)
        event_sink = Mock()
        doge = DataOnlineGenerator([STATE_1, STATE_2],
                                   STATE_1,
                                   factory,
                                   1, 1000, 2)
        doge.add_transition(TRANSITION_1, STATE_1, STATE_2, 100, action_callback, [event_sink])

        doge.start()

        assert action_callback.call_count == 1
        self.__assert_action_called(action_callback.call_args_list[0].args,
                                    SOME_SUBJECT_1, TRANSITION_1, STATE_1, STATE_2)
        assert event_sink.collect.call_count == 1
        self.__assert_event_sink_collect_called(event_sink.collect.call_args_list[0].args,
                                                0, SOME_SUBJECT_1, TRANSITION_1, STATE_1, STATE_2)
        event_sink.close.assert_called()

    def test_should_fail_to_transition_once(self):
        factory = self.__create_factory_mock([SOME_SUBJECT_1])
        action_callback = Mock()
        action_callback.side_effect = [False, True]
        event_sink = Mock()
        doge = DataOnlineGenerator([STATE_1, STATE_2, STATE_3],
                                   STATE_1,
                                   factory,
                                   1, 1000, 2)
        doge.add_transition(TRANSITION_1, STATE_1, STATE_2, 100, action_callback, [event_sink])
        doge.add_transition(TRANSITION_2, STATE_2, STATE_3, 100, action_callback, [event_sink])

        doge.start()

        assert action_callback.call_count == 2
        self.__assert_action_called(action_callback.call_args_list[0].args,
                                    SOME_SUBJECT_1, TRANSITION_1, STATE_1, STATE_2)
        self.__assert_action_called(action_callback.call_args_list[1].args,
                                    SOME_SUBJECT_1, TRANSITION_1, STATE_1, STATE_2)
        assert event_sink.collect.call_count == 1
        self.__assert_event_sink_collect_called(event_sink.collect.call_args_list[0].args,
                                                1000, SOME_SUBJECT_1, TRANSITION_1, STATE_1, STATE_2)
        event_sink.close.assert_called()

    def test_should_raise_value_error_when_provided_incorrect_state_when_adding_transition(self):
        factory = self.__create_factory_mock([])
        doge = DataOnlineGenerator([STATE_1, STATE_2, STATE_3],
                                   STATE_1,
                                   factory,
                                   1, 1000, 2)

        with pytest.raises(ValueError):
            doge.add_transition(TRANSITION_1, 'unknown_state', STATE_2, 100)
        with pytest.raises(ValueError):
            doge.add_transition(TRANSITION_1, STATE_1, 'unknown_state', 100)

    def test_should_raise_value_error_when_provided_incorrect_initial_state(self):
        factory = self.__create_factory_mock([])
        with pytest.raises(ValueError):
            DataOnlineGenerator([STATE_1, STATE_2, STATE_3],
                                'unknown_state',
                                factory,
                                1, 1000, 2)

    @staticmethod
    def __create_factory_mock(subjects):
        factory = Mock(SubjectFactory)
        factory.create = Mock()
        factory.create.side_effect = subjects
        return factory

    @staticmethod
    def __assert_action_called(args: Tuple, subject: SomeSubject, trigger: str, from_state: str, to_state: str):
        assert args[0] == subject
        assert args[1].trigger == trigger
        assert args[1].from_state == from_state
        assert args[1].to_state == to_state

    @staticmethod
    def __assert_event_sink_collect_called(args: Tuple,
                                           timestamp: int,
                                           subject: SomeSubject,
                                           trigger: str,
                                           from_state: str,
                                           to_state: str):
        assert args[0] == timestamp
        assert args[1] == subject
        assert args[2].trigger == trigger
        assert args[2].from_state == from_state
        assert args[2].to_state == to_state



from typing import Dict, Any
from unittest.mock import Mock

import pytest
from confluent_kafka import DeserializingConsumer
from confluent_kafka.admin import AdminClient
from confluent_kafka.cimpl import NewTopic
from confluent_kafka.schema_registry import SchemaRegistryClient
from confluent_kafka.schema_registry.avro import AvroDeserializer
from testcontainers.kafka import KafkaContainer

from doge_datagen import KafkaAvroSinkFactory, Subject, Transition
from tests.testcontainers.schemaregistry import SchemaRegistryContainer

TEST_TOPIC = 'test-topic'
KEY_SCHEMA = """
{
  "namespace": "com.getindata.doge",
  "type": "record",
  "name": "SomeKey",
  "fields": [
    { "name":  "key", "type":  "string"}
  ]
}
"""
VALUE_SCHEMA = """
{
  "namespace": "com.getindata.doge",
  "type": "record",
  "name": "SomeValue",
  "fields": [
    { "name":  "value", "type":  "string"}
  ]
}
"""


def key_function(subject: Subject, transition: Transition) -> Dict[str, Any]:
    return {'key': 'some_key'}


def value_function(timestamp: int, subject: Subject, transition: Transition) -> Dict[str, Any]:
    return {'value': 'some_value'}


@pytest.fixture
def kafka():
    with KafkaContainer() as kafka:
        yield kafka


@pytest.fixture
def schema_registry(kafka):
    with SchemaRegistryContainer.with_kafka_container(kafka) as schema_registry:
        yield schema_registry


class TestKafkaAvroSink:
    def test_kafka_avro_sink(self, kafka, schema_registry):
        self.__create_test_topic(kafka.get_bootstrap_server())
        self.__test_kafka_avro_sink(kafka.get_bootstrap_server(), schema_registry.get_url())

    def __test_kafka_avro_sink(self, kafka_bootstrap, schema_registry_url):
        factory = KafkaAvroSinkFactory([kafka_bootstrap], schema_registry_url, 'test-client')
        sink = factory.create(TEST_TOPIC, key_function, KEY_SCHEMA, value_function, VALUE_SCHEMA)

        sink.collect(123, Mock(), Mock())
        sink.close()

        consumer = self.__create_consumer(kafka_bootstrap, schema_registry_url)
        consumer.subscribe([TEST_TOPIC])
        msg = consumer.poll(10)
        assert msg.key()['key'] == 'some_key'
        assert msg.value()['value'] == 'some_value'

    @staticmethod
    def __create_consumer(kafka_bootstrap, schema_registry_url):
        deserializer = AvroDeserializer(SchemaRegistryClient({'url': schema_registry_url}))
        consumer = DeserializingConsumer({
            'bootstrap.servers': kafka_bootstrap,
            'group.id': 'test-group',
            'auto.offset.reset': 'earliest',
            'key.deserializer': deserializer,
            'value.deserializer': deserializer})
        return consumer

    @staticmethod
    def __create_test_topic(kafka_bootstrap):
        admin_client = AdminClient({"bootstrap.servers": kafka_bootstrap})
        admin_client.create_topics([NewTopic(TEST_TOPIC, 1, 1)])[TEST_TOPIC].result(10)



from typing import Iterable, Callable, Any, Dict

from confluent_kafka.schema_registry import SchemaRegistryClient
from confluent_kafka.schema_registry.avro import AvroSerializer

from doge_datagen import Subject, Transition, KafkaSinkFactory, KafkaSink


class KafkaAvroSinkFactory(object):

    def __init__(self,
                 bootstrap_servers: Iterable[str],
                 schema_registry_url: str,
                 client_id: str,
                 buffer_size=100000):
        """
        :param bootstrap_servers: list of bootstrap servers
        :type bootstrap_servers: Iterable[str]
        :param schema_registry_url: schema registry url for example http://localhost:8081
        :type schema_registry_url: str
        :param client_id: sink client id
        :type client_id: str
        """
        self.factory = KafkaSinkFactory(bootstrap_servers, client_id, buffer_size)
        schema_registry_conf = {'url': schema_registry_url}
        self.schema_registry_client = SchemaRegistryClient(schema_registry_conf)

    def create(self, topic: str,
               key_function: Callable[[Subject, Transition], Dict[str, Any]],
               key_schema: str,
               value_function: Callable[[int, Subject, Transition], Dict[str, Any]],
               value_schema: str) -> KafkaSink:
        """
        :param topic: topic name to which events will be emitted
        :type topic: str
        :param key_function: function that converts subject and transition to a format consumable by key serializer
        :type key_function: Callable[[Subject, Transition], K]
        :param key_schema: Avro compliant schema for key serialization
        :type key_schema: str
        :param value_function: function that converts timestamp, subject and transition to a format consumable by
            value serializer
        :type value_function: Callable[[int, Subject, Transition], V]
        :param value_schema: Avro compliant schema for value serialization
        :type value_schema: str
        :return: KafkaSink instance
        :rtype: KafkaSink
        """
        key_serializer = AvroSerializer(self.schema_registry_client, key_schema)
        value_serializer = AvroSerializer(self.schema_registry_client, value_schema)
        return self.factory.create(topic, key_function, value_function, key_serializer, value_serializer)



from .doge import Subject
from .doge import SubjectFactory
from .doge import Transition
from .doge import DataOnlineGenerator
from .doge import EventSink

from .db_sink_factory import DbSinkFactory, DbSink
from .kafka_sink_factory import KafkaSinkFactory, KafkaSink
from .kafka_avro_sink_factory import KafkaAvroSinkFactory
from .printing_sink import PrintingSink



from typing import Dict, Any

from doge_datagen import Transition, KafkaAvroSinkFactory
from examples.doge_example_common import create_example_data_online_generator, User


def key_function(subject: User, transition: Transition) -> Dict[str, Any]:
    return {'key': str(subject.user_id)}


def value_function(timestamp: int, subject: User, transition: Transition) -> Dict[str, Any]:
    value = {
        'timestamp': timestamp,
        'user': {
            'userId': str(subject.user_id),
            'balance': str(subject.balance),
            'loanBalance': str(subject.loan_balance)
        },
        'event': transition.trigger
    }
    return value


def get_schema(schema_path):
    with open(schema_path) as f:
        return f.read()


if __name__ == '__main__':
    key_schema = get_schema('./avro/Key.avsc')
    event_schema = get_schema('./avro/Event.avsc')

    factory = KafkaAvroSinkFactory(['localhost:9092'], 'http://localhost:8081', 'doge-kafka-example')
    sink = factory.create('test_avro_topic', key_function, key_schema, value_function, event_schema)

    datagen = create_example_data_online_generator(sink)

    datagen.start()



from typing import Callable

from doge_datagen import Subject, Transition, EventSink


class PrintingSink(EventSink):
    def __init__(self, format_function: Callable[[int, Subject, Transition], str]):
        self.format_function = format_function

    def collect(self, timestamp: int, subject: Subject, transition: 'Transition'):
        print(self.format_function(timestamp, subject, transition))



import os
from typing import Dict, Any

from doge_datagen import Transition, DbSinkFactory
from examples.doge_example_common import create_example_data_online_generator, User


def row_mapper_function(timestamp: int, subject: User, transition: Transition) -> Dict[str, Any]:
    row = {
        'timestamp': timestamp,
        'user_id': subject.user_id,
        'balance': subject.balance,
        'loan_balance': subject.loan_balance,
        'event': transition.trigger
    }
    return row


if __name__ == '__main__':
    db_pass = os.getenv('PGPASSWORD', 'postgres')
    factory = DbSinkFactory('postgresql://postgres:{}@localhost:5432/postgres'.format(db_pass))
    sink = factory.create('events', row_mapper_function)

    datagen = create_example_data_online_generator(sink)

    datagen.start()



import os
from typing import Dict, Any

from doge_datagen import DataOnlineGenerator, KafkaAvroSinkFactory, DbSinkFactory, Subject, Transition

from examples.doge_example_common import income_callback, spending_callback, take_loan_callback, UserFactory, User
from examples.doge_kafka_avro_example import key_function, get_schema

# Common
kafka_avro_factory = KafkaAvroSinkFactory(['localhost:9092'], 'http://localhost:8081', 'doge-demo')
db_pass = os.getenv('PGPASSWORD', 'postgres')
db_factory = DbSinkFactory('postgresql://postgres:{}@localhost:5432/postgres'.format(db_pass))
key_schema = get_schema('./avro/Key.avsc')

# TRX definition
trx_event_schema = get_schema('./avro/Trx.avsc')


def trx_value_function(timestamp: int, subject: User, transition: Transition) -> Dict[str, Any]:
    value = {
        'timestamp': timestamp,
        'user_id': str(subject.user_id),
        'amount': subject.last_payment
    }
    return value


trx_sink = kafka_avro_factory.create("trx", key_function, key_schema, trx_value_function, trx_event_schema)


# Clickstream definition
clickstream_event_schema = get_schema('./avro/Clickstream.avsc')


def clickstream_value_function(timestamp: int, subject: User, transition: Transition) -> Dict[str, Any]:
    value = {
        'timestamp': timestamp,
        'user_id': str(subject.user_id),
        'event': transition.trigger
    }
    return value


clickstream_sink = kafka_avro_factory.create("clickstream", key_function, key_schema,
                                             clickstream_value_function, clickstream_event_schema)


# Balance definition
def balance_mapper_function(timestamp: int, subject: User, transition: Transition) -> Dict[str, Any]:
    row = {
        'timestamp': timestamp,
        'user_id': str(subject.user_id),
        'balance': subject.balance,
    }
    return row


balance_sink = db_factory.create("balance", balance_mapper_function)


# Loan definition
def loan_mapper_function(timestamp: int, subject: User, transition: Transition) -> Dict[str, Any]:
    row = {
        'timestamp': timestamp,
        'user_id': str(subject.user_id),
        'loan': subject.loan_balance,
    }
    return row


loan_sink = db_factory.create("loan", loan_mapper_function)


# Doge configuration
def create_example_data_online_generator():
    datagen = DataOnlineGenerator(['offline', 'online', 'loan_screen'], 'offline', UserFactory(), 10, 60000, 1000)
    datagen.add_transition('income', 'offline', 'offline', 0.01,
                           action_callback=income_callback, event_sinks=[balance_sink])
    datagen.add_transition('spending', 'offline', 'offline', 0.1,
                           action_callback=spending_callback, event_sinks=[trx_sink, balance_sink])
    datagen.add_transition('login', 'offline', 'online', 0.1, event_sinks=[clickstream_sink])
    datagen.add_transition('logout', 'online', 'offline', 70, event_sinks=[])
    datagen.add_transition('open_loan_screen ', 'online', 'loan_screen', 30, event_sinks=[clickstream_sink])
    datagen.add_transition('close_loan_screen', 'loan_screen', 'online', 40, event_sinks=[clickstream_sink])
    datagen.add_transition('take_loan', 'loan_screen', 'online', 10,
                           action_callback=take_loan_callback, event_sinks=[clickstream_sink, loan_sink, balance_sink])
    return datagen


if __name__ == '__main__':
    datagen = create_example_data_online_generator()
    datagen.start()



from typing import Dict, Any
from unittest.mock import Mock

import pytest
import sqlalchemy
from testcontainers.postgres import PostgresContainer

from doge_datagen import DbSinkFactory, Subject, Transition


def row_mapper_function(timestamp: int, subject: Subject, transition: Transition) -> Dict[str, Any]:
    return {'some_column': 'some_value'}


@pytest.fixture
def postgres():
    with PostgresContainer() as postgres:
        yield postgres


@pytest.fixture
def engine(postgres):
    engine = sqlalchemy.create_engine(postgres.get_connection_url())
    engine.execute('create table some_table (some_column varchar(20))')
    yield engine


class TestDbSink:
        
    def test_db_sink(self, postgres, engine):
        factory = DbSinkFactory(postgres.get_connection_url())
        sink = factory.create('some_table', row_mapper_function)

        sink.collect(123, Mock(), Mock())
        sink.close()

        row = engine.execute("select * from some_table").fetchone()
        assert row['some_column'] == 'some_value'

    def test_table_exist_validation(self, postgres):
        factory = DbSinkFactory(postgres.get_connection_url())
        with pytest.raises(ValueError):
            factory.create('some_other_table', row_mapper_function)



import logging
import os
import time

from abc import ABC, abstractmethod
from datetime import datetime
from random import random
from typing import List, Dict, Callable, Iterable, Set, TypeVar, Generic


Subject = TypeVar('Subject')

logging.basicConfig(level=os.environ.get("LOGLEVEL", "INFO"))
LOGGER = logging.getLogger(__name__)


class SubjectFactory(Generic[Subject]):
    """
    SubjectFactory creates instances of Subject class. In order to provide attributes to Subject class it is
    expected that SubjectFactory will be extended.
    """
    @abstractmethod
    def create(self) -> Subject:
        """
        Subject instance represents a single item that traverse defined state machine. It also can be used to hold data
        that can be modified and emitted during transitions. On its own it provides no methods nor attributes.
        It is expected that attributes will be provided by SubjectFactory during creation time.
        :return: a new Subject instance
        :rtype: Subject
        """
        pass


class EventSink(ABC):
    """
    Abstract base class for Sinks that will consume events emitted during state machine transitions.
    """
    @abstractmethod
    def collect(self, timestamp: int, subject: Subject, transition: 'Transition'):
        """
        Executed during state machine transition.
        :param timestamp: event timestamp
        :type timestamp: int
        :param subject: subject instance that makes a transition
        :type subject: Subject
        :param transition: transition instance on which subject travels
        :type transition: Transition
        """
        pass

    def close(self):
        """
        Executed at the end of DataOnlineGenerator.start() method execution. This method allows for graceful flush
        and close of any resource that was used as event sink
        """
        pass


class Transition(object):
    """
    Represents a possible transition in state machine.
    """
    def __init__(self,
                 trigger: str,
                 from_state: str,
                 to_state: str,
                 probability: float,
                 action_callback: Callable[[Subject, 'Transition'], bool] = None,
                 event_sinks: List[EventSink] = None):
        self.trigger = trigger
        self.from_state = from_state
        self.to_state = to_state
        self.probability = probability
        self.action_callback = action_callback
        self.event_sinks = event_sinks or ()


class DataOnlineGenerator(Generic[Subject]):
    """
    DataOnlineGenerator is a state machine which is traversed by multiple subjects automatically based on probability
    defined for each possible transition.

    State machine works in ticks which length and number is defined in constructor. In each tick DataOnlineGenerate
    evaluates each subject and makes a transition based on provided probabilities. Probabilities of doing a transition
    from given state have to be less or equal to 100. If probabilities of all transitions are less than 100 remaining
    value is treated as probability of staying in the same state in given tick.
    """
    STAY_TRIGGER = "__stay"
    transition_matrix: Dict[str, Dict[str, Transition]]
    probability_matrix: Dict[str, Dict[str, float]]
    subjects_states: Dict[Subject, str]
    sinks: Set[EventSink]

    def __init__(self,
                 states: List[str],
                 initial_state: str,
                 subject_factory: SubjectFactory,
                 subjects_num: int,
                 tick_ms: int,
                 ticks_num: int,
                 timestamp_start: int = 0):
        """
        :param states: list of all states in state machine
        :type states: List[str]
        :param initial_state: initial state for all subjects
        :type initial_state: str
        :param subject_factory: subject factory instance which will be used to create all subjects
        :type subject_factory: SubjectFactory
        :param subjects_num: number of subjects to create
        :type subjects_num: int
        :param tick_ms: length of a tick. It is assumed it is in milliseconds. It will be used to advance timestamp.
        :type tick_ms: int
        :param ticks_num: number of ticks to simulate
        :type ticks_num: int
        :param timestamp_start: optional starting timestamp
        :type timestamp_start: int
        """
        self.states = states
        self.__validate_state_defined(initial_state)
        self.initial_state = initial_state
        self.subject_factory = subject_factory
        self.subjects_num = subjects_num
        self.tick_ms = tick_ms
        self.ticks_num = ticks_num
        self.timestamp = timestamp_start
        self.__init_matrices()
        self.subjects_states = {}
        self.sinks = set()

    def __init_matrices(self):
        self.transition_matrix = {}
        self.probability_matrix = {}
        for state in self.states:
            self.transition_matrix[state] = {}
            self.probability_matrix[state] = {}

    def add_transition(self,
                       trigger: str,
                       from_state: str,
                       to_state: str,
                       probability: float,
                       action_callback: Callable[[Subject, 'Transition'], bool] = None,
                       event_sinks: List[EventSink] = None):
        """
        :param trigger: name of the trigger that triggers this transition. Has to be unique across all transitions
            from state :param from_state
        :type trigger: str
        :param from_state: name of the state transition points from
        :type from_state: str
        :param to_state: name of the state transition points to
        :type to_state: str
        :param probability: probability (in percentage, where 100 means certain probability) of Subject choosing this
            transition when being in state :param from_state
        :type probability: float
        :param action_callback: optional function that accepts Subject and Transition and returns a boolean value.
            It allows subject state manipulation and its return value is used as a condition for transition. If False
            is returned subject will stay in current state, transition will be not made and no event will be emitted.
        :type action_callback: Callable[[Subject, 'Transition'], bool]
        :param event_sinks: optional instance of EventSink class. During transition event_sink method collect will
            be called.
        :type event_sinks: EventSink
        """
        self.__validate_state_defined(from_state)
        self.__validate_state_defined(to_state)

        if event_sinks:
            self.sinks.update(event_sinks)

        transition = Transition(trigger, from_state, to_state, probability, action_callback, event_sinks)

        self.transition_matrix[from_state][trigger] = transition
        self.probability_matrix[from_state][trigger] = probability

        self.__validate_probability_sum(from_state)

    def __add_stay_transitions(self):
        for state in self.states:
            transitions = self.transition_matrix[state].values()
            if len(transitions) == 0:
                LOGGER.warning("State {} is a final state and has no output transitions".format(state))
            stay_probability = 100 - self.__get_probability_sum(transitions)
            self.transition_matrix[state][self.STAY_TRIGGER] = \
                Transition(self.STAY_TRIGGER, state, state, stay_probability)
            self.probability_matrix[state][self.STAY_TRIGGER] = stay_probability

    def __validate_probability_sum(self, state: str):
        transitions = self.transition_matrix[state].values()
        probability_sum = self.__get_probability_sum(transitions)
        if probability_sum > 100:
            raise ValueError("Sum of probabilities for state {} have exceeded 100%!".format(state))

    def __validate_state_defined(self, state: str):
        if state not in self.states:
            raise ValueError('State {} is not present in machine states {}'.format(state, self.states))

    @staticmethod
    def __get_probability_sum(transitions: Iterable[Transition]) -> float:
        return sum(map(lambda transition: transition.probability, transitions))

    def __generate_subjects(self):
        for i in range(self.subjects_num):
            self.subjects_states[self.subject_factory.create()] = self.initial_state

    def __tick(self):
        for subject, state in self.subjects_states.items():
            trigger = self.__random_trigger(state)
            transition = self.transition_matrix[state][trigger]

            if not trigger == self.STAY_TRIGGER:
                should_transition = True
                if transition.action_callback:
                    should_transition = transition.action_callback(subject, transition)
                if should_transition:
                    self.subjects_states[subject] = transition.to_state
                    for sink in transition.event_sinks:
                        sink.collect(self.timestamp, subject, transition)

        self.timestamp += self.tick_ms

    # this is faster than using random.choice()
    def __random_trigger(self, state: str) -> str:
        rand = random() * 100
        cum = 0
        for key, probability in self.probability_matrix[state].items():
            cum += probability
            if rand < cum:
                return key

    def __close_sinks(self):
        for sink in self.sinks:
            sink.close()

    def start(self):
        start = time.time()
        LOGGER.info("Data generation start: {}".format(datetime.now()))
        self.__generate_subjects()
        self.__add_stay_transitions()
        for i in range(self.ticks_num):
            self.__tick()
        self.__close_sinks()
        LOGGER.info("Data generation finished: {}".format(datetime.now()))
        end = time.time()
        LOGGER.info("Execution took {} s".format(end - start))



from dataclasses import dataclass
from random import randrange

from doge_datagen import Transition, SubjectFactory, DataOnlineGenerator


@dataclass
class User:
    user_id: int
    balance: int
    loan_balance: int
    last_payment: int

    def __hash__(self):
        return self.user_id


class UserFactory(SubjectFactory[User]):
    def __init__(self):
        super().__init__()
        self.current_id = 0

    def create(self) -> User:
        user = User(self.current_id, randrange(0, 1000), 0, 0)
        self.current_id += 1
        return user


def income_callback(subject: User, transition: Transition):
    subject.balance += randrange(500, 2000)
    return True


def spending_callback(subject: User, transition: Transition):
    payment = randrange(10, 100)
    if payment > subject.balance:
        return False
    subject.balance -= payment
    subject.last_payment = payment
    return True


def take_loan_callback(subject: User, transition: Transition):
    subject.loan_balance += 10000
    subject.balance += 10000
    return True


def create_example_data_online_generator(sink):
    datagen = DataOnlineGenerator(['offline', 'online', 'loan_screen'], 'offline', UserFactory(), 10, 60000, 1000)
    datagen.add_transition('income', 'offline', 'offline', 0.01, action_callback=income_callback, event_sinks=[sink])
    datagen.add_transition('spending', 'offline', 'offline', 0.1, action_callback=spending_callback, event_sinks=[sink])
    datagen.add_transition('login', 'offline', 'online', 0.1, event_sinks=[sink])
    datagen.add_transition('logout', 'online', 'offline', 70, event_sinks=[sink])
    datagen.add_transition('open_loan_screen', 'online', 'loan_screen', 30, event_sinks=[sink])
    datagen.add_transition('close_loan_screen', 'loan_screen', 'online', 40, event_sinks=[sink])
    datagen.add_transition('take_loan', 'loan_screen', 'online', 10,
                           action_callback=take_loan_callback, event_sinks=[sink])
    return datagen



from doge_datagen import PrintingSink, Transition
from examples.doge_example_common import create_example_data_online_generator, User


def format_function(timestamp: int, user: User, transition: Transition) -> str:
    return '[{}] User id: {}, balance: {}, loan_balance: {} made a transition {} from {} to {}'\
        .format(timestamp,
                user.user_id,
                user.balance,
                user.loan_balance,
                transition.trigger,
                transition.from_state,
                transition.to_state)


if __name__ == '__main__':
    sink = PrintingSink(format_function)

    datagen = create_example_data_online_generator(sink)

    datagen.start()



import os
import sys

from setuptools import setup
from setuptools import find_packages
from setuptools.command.install import install


__version__ = "1.0.4"

with open("README.md", "r") as fh:
    long_description = fh.read()


def get_requirements(filename):
    with open(filename, "r", encoding="utf-8") as fp:
        reqs = [x.strip() for x in fp.read().splitlines()
                if not x.strip().startswith('#') and not x.strip().startswith('-i')]
    return reqs


class VerifyVersionCommand(install):
    """Custom command to verify that the git tag matches our version"""
    description = 'verify that the git tag matches our version'

    def run(self):
        tag = os.getenv('CI_COMMIT_TAG')

        if tag != f"v{__version__}":
            info = "Git tag: {0} does not match the version of this app: {1}".format(
                tag, __version__
            )
            sys.exit(info)


setup(
    name="doge_datagen",
    version=__version__,
    author="GetInData",
    author_email="office@getindata.com",
    description="Data Online Generator",
    long_description=long_description,
    long_description_content_type="text/markdown",
    license="Apache Software License (Apache 2.0)",
    url="https://github.com/getindata/doge-datagen",
    packages=find_packages(),
    python_requires='>=3.8',
    classifiers=[
        "Programming Language :: Python :: 3.8",
        "Operating System :: OS Independent",
    ],
    install_requires=get_requirements('requirements.txt'),
    py_modules=['doge_datagen'],
    cmdclass={
        'verify': VerifyVersionCommand,
    }
)



import json

from doge_datagen import Transition, KafkaSinkFactory
from examples.doge_example_common import create_example_data_online_generator, User


def key_function(subject: User, transition: Transition) -> str:
    return str(subject.user_id)


def value_function(timestamp: int, subject: User, transition: Transition) -> str:
    value = {
        'timestamp': timestamp,
        'user': {
            'user_id': subject.user_id,
            'balance': subject.balance,
            'loan_balance': subject.loan_balance
        },
        'event': transition.trigger
    }
    return json.dumps(value)


if __name__ == '__main__':
    factory = KafkaSinkFactory(['localhost:9092'], 'doge-kafka-example')
    sink = factory.create('test_topic', key_function, value_function)

    datagen = create_example_data_online_generator(sink)

    datagen.start()



import sys
from typing import Iterable, Callable, TypeVar

from confluent_kafka import Producer, Message, KafkaError
from confluent_kafka.serialization import SerializationContext, StringSerializer, Serializer, MessageField

from doge_datagen import Subject, Transition, EventSink

K = TypeVar('K')
V = TypeVar('V')


class CounterHolder:
    def __init__(self):
        self.counter = 0


class KafkaSink(EventSink):
    def __init__(self,
                 producer: Producer,
                 msg_count_holder: CounterHolder,
                 topic: str,
                 key_function: Callable[[Subject, Transition], K],
                 key_serializer: Serializer,
                 value_function: Callable[[int, Subject, Transition], V],
                 value_serializer: Serializer,
                 buffer_size=100000):
        self.producer = producer
        self.msg_count_holder = msg_count_holder
        self.topic = topic
        self.key_function = key_function
        self.key_serializer = key_serializer
        self.value_function = value_function
        self.value_serializer = value_serializer
        self.buffer_size = buffer_size

    @staticmethod
    def __on_delivery(err: KafkaError, msg: Message):
        if err:
            print(str(err), file=sys.stderr)

    def collect(self, timestamp: int, subject: Subject, transition: 'Transition'):

        key_ctx = SerializationContext(self.topic, MessageField.KEY)
        key = self.key_serializer(self.key_function(subject, transition), key_ctx)
        value_ctx = SerializationContext(self.topic, MessageField.VALUE)
        value = self.value_serializer(self.value_function(timestamp, subject, transition), value_ctx)
        if self.msg_count_holder.counter >= self.buffer_size:
            self.producer.flush()
            self.msg_count_holder.counter = 0
        self.producer.produce(self.topic,
                              key=key,
                              value=value,
                              timestamp=timestamp,
                              on_delivery=self.__on_delivery)
        self.msg_count_holder.counter += 1

    def close(self):
        self.producer.flush()


class KafkaSinkFactory(object):
    def __init__(self,
                 bootstrap_servers: Iterable[str],
                 client_id: str,
                 buffer_size=100000):
        """
        :param bootstrap_servers: list of bootstrap servers
        :type bootstrap_servers: Iterable[str]
        :param client_id: sink client id
        :type client_id: str
        """
        conf = {
            'bootstrap.servers': ','.join(bootstrap_servers),
            'client.id': client_id,
        }
        self.producer = Producer(conf)
        self.msg_count = 0
        self.buffer_size = buffer_size
        self.counter_holder = CounterHolder()

    def create(self, topic: str,
               key_function: Callable[[Subject, Transition], K],
               value_function: Callable[[int, Subject, Transition], V],
               key_serializer=StringSerializer(),
               value_serializer=StringSerializer()) -> KafkaSink:
        """
        :param topic: topic name to which events will be emitted
        :type topic: str
        :param key_function: function that converts subject and transition to a format consumable by key serializer
        :type key_function: Callable[[Subject, Transition], K]
        :param value_function: function that converts timestamp, subject and transition to a format consumable by
            value serializer
        :type value_function: Callable[[int, Subject, Transition], V]
        :param key_serializer: Serializer instance that will be used to serialize key_function output into bytes
        :type key_serializer: Serializer
        :param value_serializer: Serializer instance that will be used to serialize value_function output into bytes
        :type value_serializer: Serializer
        :return: KafkaSink instance
        :rtype: KafkaSink
        """
        return KafkaSink(self.producer,
                         self.counter_holder,
                         topic,
                         key_function,
                         key_serializer,
                         value_function,
                         value_serializer,
                         self.buffer_size)



from typing import Callable, Dict, Any, List

from sqlalchemy import create_engine, MetaData, Table
from sqlalchemy.engine import Engine

from doge_datagen import Subject, Transition
from doge_datagen import EventSink


class DbSink(EventSink):
    def __init__(self,
                 engine: Engine,
                 metadata: MetaData,
                 table_name: str,
                 row_mapper_function: Callable[[int, Subject, Transition], Dict[str, Any]],
                 batch_size: int):
        self.engine = engine
        self.table = Table(table_name, metadata)
        self.batch: List[Dict[str, Any]] = []
        self.row_mapper_function = row_mapper_function
        self.batch_size = batch_size
        self.__validate_table_exists()

    def collect(self, timestamp: int, subject: Subject, transition: 'Transition'):
        row = self.row_mapper_function(timestamp, subject, transition)
        self.batch.append(row)
        if len(self.batch) >= self.batch_size:
            self.__insert_batch()

    def close(self):
        if len(self.batch):
            self.__insert_batch()

    def __insert_batch(self):
        self.engine.execute(self.table.insert(), self.batch)
        self.batch = []

    def __validate_table_exists(self):
        if not self.table.exists(bind=self.engine):
            raise ValueError("Table {} does not exist".format(self.table.name))


class DbSinkFactory(object):
    batches: Dict[str, List[Dict[str, Any]]]
    tables: Dict[str, Table]

    def __init__(self, db_url: str):
        """
        :param db_url: database url in format acceptable by SQLAlchemy
        :type db_url: str
        """
        self.engine = create_engine(db_url)
        self.metadata = MetaData()
        self.metadata.reflect(self.engine)
        self.batches = {}
        self.tables = {}

    def create(self,
               table_name: str,
               row_mapper_function: Callable[[int, Subject, Transition], Dict[str, Any]],
               batch_size: int = 1000) -> DbSink:
        """
        :param table_name: name of the table that will store events
        :type table_name: str
        :param row_mapper_function: function that will map timestamp, subject and transition instance to database row
        :type row_mapper_function: Callable[[int, Subject, Transition], Dict[str, Any]]
        :param batch_size: size of a batch that will be inserted to database
        :type batch_size: int
        :return: DbSink instance
        :rtype: DbSink
        """
        return DbSink(self.engine, self.metadata, table_name, row_mapper_function, batch_size)
